 #
 # Pcom: a personal communication shell
 # by: Theo Verelst (theover@yahoo.com)
 #

 # during debug:
 #set t [winfo children .]
if [winfo exists .pc] {
 set t {.pc.f .pc.fl}
 foreach i [winfo children .pc]  {
   if {[winfo toplevel $i] != "$i"} {
      lappend t $i
   }
 }
 foreach i $t {destroy $i}
 update
} {
 toplevel .pc
}
wm title .pc "Pcom Local"

 set myss -1      ;# my server socket
 set rems -1      ;# remote socket
 set myip -1      ;# local IP address
 set myport 4404   ;# default port

 set state ready
 set allowtcl 1
                  # quite dangerous, a total security loophole...
                  # set to zero when in doubt...
                  # it allows remote command execution
 set icanconnect 0

 set crdir /
 set crpat *
 set crdirs {}
 set crfiles {}
 set cldir /
 set clpat *

 # I don't think this works anymore..
 proc forceclose {{max 30} {from 0} {type sock}} {
   log "forced closed: "
   for {set i 0} {$i <30} {incr i}  {
      set name "$type[expr $from + $i]"
      if {[catch "close $name"] == 0} {
         log " $name"
      }
   }
   log \n
 }

 proc log {l} {
   if {[winfo exists .pc.tlog]} {
      .pc.tlog insert end $l
      .pc.tlog see end
   } else {
      puts -nonewline "$l"
   }
 }

 # Incoming lines are parsed based on the first word, the command,
 # and the second 'argument' or the rest of the line, which in some cases
 # gets parsed further.

 proc parse_in {in} {
   global state allowtcl

 #   set com [lindex $in 0]
   set com [string range $in 0 [expr [string first " " $in] -1] ]
   set rest [string range $in [expr [string first " " $in] +1] end ]
   switch $com \
   {DO_LS} {
      set olddir [pwd]; set dirs {}; set files {}
      set dir [lindex $rest 0]; set pattern [lindex $rest 1]
      if {[catch {cd $dir}] != 0} {
         pcsend "VALUE_LS $dir $pattern [list {non-existent dir} {}]"
         cd $olddir
         log "DO_LS $dir $pattern (Error: non-existent dir)\n"
         return
      }
      puts [pwd]
      foreach i [glob $pattern] {
         if [file exists $i] {
            if [file isdir $i] {
               lappend dirs $i
            } else {
               lappend files $i
            }
         }
      }
      pcsend "VALUE_LS $dir $pattern [list $dirs $files]"
      cd $olddir
      log "DO_LS $dir $pattern\n"
   } \
   {DO_TCL} {
      if {$allowtcl != 0} {
         set tcl $rest
         set cr [catch $tcl tclret]
 #         if {$cr == 0} {
 #            if {[info exists tclret] == 0} {set tclret ""}
 #         } else {
 #            set tclret ERROR
 #        }
         log "Tcl executed:\n$tcl\n"
         pcsend "DO_TCLRET $tclret"
      } else {
         log "Tcl command blocked (not allowed):\n$tcl\n"
      }
   } \
   {DO_SETVAR} {
      if {$allowtcl != 0} {
         set tclvar [lindex $rest 0]
         set tclval [lrange $rest 1 end]
         global $tclvar
         eval set $tclvar $tclval
 #         puts $tclvar
         log "Tcl set executed: set $tclvar $tclval\n"
      } else {
         log "Tcl set command blocked (not allowed): set $tclvar $tclval\n"
      }
   } \
   {DO_TCLRET} {
      if {$rest != ""} {log "Tcl Return value: $rest\n"}
      set state ready
   } \
   {VALUE_LS} {
      global crdir crpat crdirs crfiles
      set crdir   [lindex $rest 0]
      set crpat   [lindex $rest 1]
      set crdirs  [lindex $rest 2]
      set crfiles [lindex $rest 3]
      # puts "$rest\n$crdirs,$crfiles"
      .pc.f.ld del 0 end ; .pc.f.lf del 0 end
      foreach i $crdirs {.pc.f.ld insert end $i}
      foreach i $crfiles {.pc.f.lf insert end $i}
   } \
   {DO_RECFILE} {
      log "(requested: DO_RECFILE $rest)"
      eval receive_file $rest
      pcsend "READY_RECFILE "
      log "READY_RECFILE\n"
   } \
   {DO_SENDFILE} {
      log "(requested: DO_SENDFILE $rest)"
      eval send_file $rest
      pcsend "READY_SENDFILE "
      log "READY_SENDFILE\n"
   } \
   {DO_REQSENDFILE} {
      pcsend "DO_SENDFILE $rest"
      log "DO_SENDFILE $rest\n"
   } \
   {DO_ABORTFILE} {
      catch "close $filess; close $fileso; close $filed"
      log "Aborted file transfer.\n"
   } \
   {DO_CLOSECONTROL} {
      global rems
      disconnect $rems
      set rems -1
      log "Closed control connection.\n"
   } \
   {default} {
      bell
      .pc.tcom insert end "REMOTE: $in\n"
      .pc.tcom see end
      set state ready
      after 200 bell
   }
 }

 proc serv {sock ip t} {
   global rems remip
   log "client connected, socket $sock, ip adr $ip\n"
   if {$rems != -1} {
      log "client connect attempt while already connected, ignored\n"
      close $sock
      return
   }
     set t1 [lindex [split $ip .] 0]
     set t2 [lindex [split $ip .] 1]
     if {!($t1 == 192 && $t2 == 168) && !($t1 == 127 && $t2 == 0) } {
        log "connection attempt from illegal IP address $ip for this Pcom\n"
        close $sock
        return
     }

   set rems $sock
   set remip $ip
   fileevent $rems readable {
      global rems
      set in [gets $rems]
      parse_in $in
      if [eof $rems] {
         disconnect $rems
	 set rems -1
      }
    }
 }

 proc init {} {
   global myport myip myss
 # figure out ip address before server socket is started
   myip
 # log window
   text .pc.tlog -width 40 -height 4
   pack  .pc.tlog -side bottom -anchor s -expand n -fill both

   log "my IP address is $myip.\n"
 #  set up a listening socket
   catch {set myss [socket -server serv $myport]}
   if {$myss != -1} {
      log "listening with $myss at port $myport\n"
   } else {
      log "server socket already in use, use active connect.\n"
   }

   log "Init OK.\n"
 }

 proc tserv {sock ip t} {   # simply to figure out IP address
   global ts
   set ts $sock
 }

 proc myip {{port 3020}} {    # Figure out what this machine's IP address is.
   global myip
   set tss [socket -server tserv $port]
   set ts2 [socket [info hostname] $port]
   set myip [lindex [fconfigure $ts2 -sockname] 0]
   close $tss
   close $ts2
   if [info exists ts] {close $ts}
   return $myip
 }

 proc disconnect {s} {
   close $s
 #   set $s -1
   log "disconnected (closed $s).\n"
 }

 proc pcconnect {{toip {}} {toport {}}} {
   global rems remip myport icanconnect
   if {$rems != -1} {
      log "Connect attempt while already connected: ignored"
      return
      disconnect $rems
      set rems -1
   }
   if {[catch {set rems [socket $remip $myport]}] !=0} {
      set rems -1
      log "Attempt to connect to $remip failed.\n"
      return
   }
   fileevent $rems readable {
      global rems
      set in [gets $rems]
      parse_in $in
   }
   set icanconnect 1  ;# appearently we could initiate a connection with the addressee
 }

 proc pcsend {{line \n}} {
   global rems
   if {$rems == -1} {return}
   puts $rems $line
   flush $rems
 }

 proc new_serv_address {} {
   global myss myport rems
   if {$rems != -1} {
      log "Attempt to change server IP address while already connected, ignored\n"
      return
   }
   if {$myss != -1} {
      close $myss
      log "Closed server socket $myss\n"
      set myss -1
   }
   if {[catch {set myss [socket -server serv $myport]}] == 0} {
      global icanconnect
      log "Now listening with $myss at port $myport\n"
      set icanconnect 0
   } else {
      set myss -1
      log "server address already in use, use active connect.\n"
   }
 }

 # send routines, with the ability to initiate the socket connection
 # from either the send or receive side.
 # that is, either one of two connected pcom's can initiate the connection
 # tp transfer a file a certain direction, regardless of which
 # pcom did the file transfer request. (firewall stuff))

 proc file_serv_send {s ip t} {
   global filess fileso filed
   set fileso $s
   close $filess        ;# no longer needed

    fconfigure $filed -translation binary
    fconfigure $fileso -translation binary
    set n [fcopy $filed $fileso]
    close $filed
    close $fileso
    log ", Ready ($n bytes).\n"

 }
 proc file_serv_receive {s ip t} {
   global filess fileso filed
   set fileso $s
   close $filess        ;# no longer needed

   fconfigure $filed -translation binary
   fconfigure $fileso -translation binary
   set n [fcopy $fileso $filed]
   close $fileso
   close $filed
   log ", Ready ($n bytes).\n"
 }

 proc send_file {localname port {ip {}} } {
   global filess fileport filed fileso
   if {[catch {set filed [open $localname r]}] != 0} {
      log "Unable to open file $localname to send.\n"
      return ERROR
   }
   if {$ip == {}} {
      set fileport $port
      if {[catch {set filess [socket -server file_serv_send $port]}] != 0} {
         log "Unable to open file server socket (port $port).\n"
         return ERROR
      }
      log "Transfering file $localname "
   } else {
      set fileport $port
      if {[catch {set fileso [socket $ip $port]}] != 0} {
         close $filed
         log "Unable to open connection to $ip, port $port for file tranfer.\n"
         return ERROR
      }
      fconfigure $filed -translation binary
      fconfigure $fileso -translation binary
      log "Transfering file $localname "
      set n [fcopy $filed $fileso]
      close $filed
      close $fileso
      log ", Ready ($n bytes).\n"
   }
 }

 proc receive_file {localname port {ip {}} } {
   global filess fileport filed fileso
   if {[catch {set filed [open $localname w]}] != 0} {
      log "Unable to open file $localname to receive.\n"
      return
   }
   if {$ip == {}} {
      set fileport $port
      if {[catch {set filess [socket -server file_serv_receive $port]}] != 0} {
         log "Unable to open file server socket (port $port).\n"
         return ERROR
      }
      log "Transfering file $localname "
   } else {
      set fileport $port
      if {[catch {set fileso [socket $ip $port]}] != 0} {
         close $filed
         log "Unable to open connection to $ip, port $port for file tranfer.\n"
         return ERROR
      }
      fconfigure $filed -translation binary
      fconfigure $fileso -translation binary
      log "Transfering file $localname "
      set n [fcopy $fileso $filed]
      close $fileso
      close $filed
      log ", Ready ($n bytes).\n"
   }
 }

 proc do_receive_file {rdir file ldir} {
   global myip remip icanconnect
   set recport 4305
                       # This could be any free port which is available
   if {$icanconnect == 1} {
      pcsend "DO_SENDFILE [file join $rdir $file] $recport"
      receive_file [file join $ldir $file] $recport $remip
   } else {
      receive_file [file join $ldir $file] $recport
      pcsend "DO_SENDFILE [file join $rdir $file] $recport $myip"
   }
 }

 proc do_send_file {ldir file rdir} {
   global myip remip icanconnect
   set recport 4305
                      # to get files through firewalls, use 80 ... (also see above)
   if {$icanconnect == 1} {
      pcsend "DO_RECFILE [file join $rdir $file] $recport"
      send_file [file join $ldir $file] $recport $remip
   } else {
      send_file [file join $ldir $file] $recport
      pcsend "DO_RECFILE [file join $rdir $file] $recport $myip"
   }
 }

 proc make_ui {} {
   global remip intext myport
 #   label .lip -textvar myip
 #   pack .lip -side top -anchor n -expand n -fill none
   frame .pc.fb
   pack .pc.fb -side top -anchor n -expand n -fill x
   button .pc.fb.bquit -text "Full Close" -command quit
   pack .pc.fb.bquit -side right -anchor ne -fill none -expand n
   label .pc.fb.lstate -textvar state
   pack .pc.fb.lstate -side left -anchor nw -fill none -expand n
   button .pc.fb.bna -text "New Address" -command new_serv_address
   pack .pc.fb.bna -side left -anchor nw -fill none -expand n
   button .pc.fb.bcl -text "Close Connection" -command \
       {global rems; pcsend "DO_CLOSECONTROL \n"; disconnect $rems; set rems -1 }
   pack .pc.fb.bcl -side left -anchor nw -fill none -expand n
   button .pc.fb.bcon -text Con -command {console show; puts {foreach i [.pc.f.lf get 0 end] {puts $i; update; do_receive_file $crdir $i $cldir ; update}}}
   pack .pc.fb.bcon

   frame .pc.fcl
   frame .pc.fc1
   frame .pc.fc2
   frame .pc.fc3
   pack .pc.fcl -side top -anchor n -expand n -fill x
   pack .pc.fc1 -side top -anchor n -expand n -fill x
   pack .pc.fc2 -side top -anchor n -expand n -fill x
   pack .pc.fc3 -side top -anchor n -expand n -fill x
   set remip 192.168.0.1
   entry .pc.fc1.eip -textvar remip -width 14
   entry .pc.fc2.eport -textvar myport -width 14
   entry .pc.fcl.elip -textvar myport -width 14
   label .pc.fc1.lip -text "Remote IP address:" -width 15 -anchor e
   label .pc.fc2.lport -text "Port:" -width 15 -anchor e
   label .pc.fcl.llip -text "Local server port" -width 15 -anchor e
   button .pc.fc3.connect -text Connect -command pcconnect
   pack .pc.fcl.llip -side left -expand n -fill none
   pack .pc.fcl.elip -side left -expand n -fill none
   pack .pc.fc1.lip -side left -expand n -fill none
   pack .pc.fc1.eip -side left -expand n -fill none
   pack .pc.fc2.lport -side left -expand n -fill none
   pack .pc.fc2.eport -side left -expand n -fill none
   pack .pc.fc3.connect -side left -expand y -fill x

   text .pc.tcom -width 40 -height 4
   entry .pc.ecom -textvar intext
   pack .pc.ecom -side top -anchor n -fill x -expand y
   entry .pc.ecomtcl -textvar intexttcl
   pack .pc.ecomtcl -side top -anchor n -fill x -expand y
   pack .pc.tcom -side top -anchor n -fill both -expand y
   bind .pc.ecom <Return> {
      pcsend $intext; .pc.tcom insert end "LOCAL: $intext\n";
      .pc.tcom see end
      set intext ""
   }
   bind .pc.ecomtcl <Return> {
      pcsend "DO_TCL $intexttcl"; .pc.tcom insert end "LOCALtcl: $intexttcl\n";
      .pc.tcom see end
      set intext ""
      .pc.ecomtcl sel range 0 end
   }
   # make_fileui
 }

 proc make_fileui {} {
   toplevel .pc.f
   wm title .pc.f "Remote Dir"
   listbox .pc.f.ld; listbox .pc.f.lf
   pack .pc.f.ld .pc.f.lf -side left -expand y -fill both
   entry .pc.f.ed -textvar crdir
   entry .pc.f.ep -textvar crpat
   pack .pc.f.ed .pc.f.ep
   button .pc.f.bu -text Update -command \
       {global crdir crpat; pcsend "DO_LS $crdir $crpat"}
   button .pc.f.bdu -text Up
   pack .pc.f.bu .pc.f.bdu -fill x
   .pc.f.bdu conf -command {
      set s [file split $crdir]
      if {[llength $s] > 1} {
         set up [eval file join [lrange $s  0 [expr [llength $s]-2] ] ]
         set crdir $up
      }
      .pc.f.bu invoke
   }
   bind .pc.f.ld <Double-Button-1> {
      global crdir
      set crdir [eval file join $crdir [selection get]]
      .pc.f.bu invoke
   }
   bind .pc.f.lf <Double-Button-1> {
      do_receive_file $crdir [selection get] $cldir
   }

 # local file windows
   toplevel .pc.fl
   wm title .pc.fl "Local Dir"
   listbox .pc.fl.ld; listbox .pc.fl.lf
   pack .pc.fl.ld .pc.fl.lf -side left -expand y -fill both
   entry .pc.fl.ed -textvar cldir
   entry .pc.fl.ep -textvar clpat
   pack .pc.fl.ed .pc.fl.ep
   button .pc.fl.bu -text Update -command \
       {do_local_ls }
   button .pc.fl.bdu -text Up
   pack .pc.fl.bu .pc.fl.bdu -fill x
   .pc.fl.bdu conf -command {
      set s [file split $cldir]
      if {[llength $s] > 1} {
         set up [eval file join [lrange $s  0 [expr [llength $s]-2] ] ]
         set cldir $up
      }
      .pc.fl.bu invoke
   }
   bind .pc.fl.ld <Double-Button-1> {
      global cldir
      set cldir [eval file join $cldir [selection get]]
      .pc.fl.bu invoke
   }
   bind .pc.fl.lf <Double-Button-1> {
      do_send_file $cldir [selection get] $crdir
   }
 }

 proc do_local_ls {} {
   global cldir clpat
      set olddir [pwd]; set dirs {}; set files {}
 #      puts [pwd]
      if {[catch "cd $cldir"] != 0} {
         cd $olddir
         log "Local ls: $cldir $pattern (Error: non-existent dir)\n"
      }
 #      puts [pwd]
      foreach i [glob $clpat] {
         if [file exists $i] {
            if [file isdir $i] {
               lappend dirs $i
            } else {
               lappend files $i
            }
         }
      }
      .pc.fl.ld del 0 end ; .pc.fl.lf del 0 end
      foreach i $dirs {.pc.fl.ld insert end $i}
      foreach i $files {.pc.fl.lf insert end $i}

      cd $olddir
 #      puts "$cldir $clpat {$dirs} {$files}"
 }

 proc close_all {} {
   global myss rems
   if {$myss != -1} {close $myss; set myss -1}
   if {$rems != -1} {close $rems; set rems -1}
 }

 proc quit {} {
   close_all
   log "Quit: all sockets closed.\n"
   # this was for certain particular reasons, use main window close to realy quit.
 }

 #
 # main
 #

 history keep 1000
                     # I always do this, but it makes not much sense
 # console show      # without this command of course (it seems).
 init
 make_ui
 # Comment this one out to get no file windows:
 make_fileui

