
/***** soundtest.c: socket controllable sound synthesizer code *****/

/* 
    sound lib is seperated, pa soundlib is used with OSS on linux
    contains main() and argument processing.
*/

/***********************************************************************

   COPYRIGHT       (c) Theo Verelst

  All rights are reserved by the author, who can be reached through 
      theover@yahoo.com (or see http://theover.tripod.com)
  exept that permission is granted for non-commercial use 
  of sources or programs and the sounds coming from them, 
  though the rights on the algorithms are strictly reserved.
  The author relies on the courtesy of users or programmers 
  who make small or large adaptations to inform him about them, 
  and demands that every copy or adaptation, also partial, 
  clearly contains this copyright notice.

***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include <sys/types.h>

#define PREMUL 32000           /* Fixed exponent  */
#define SRATE  44100           /* Wav Sample Rate */
#define MAXGENERATOR 256
#ifndef PI
#define PI 3.1415926535
#endif

#define GENINIT 0
#define GENRUN  1
#define GENSTOP 2
#define MAXSINTAB 262144
#define MAXSST    262144
#define MAXPOLY 32
#define MAXSAMPLETABLE 256
int       gen_state[MAXGENERATOR];     /* Generator State      */
long long gen_phase[MAXGENERATOR];     /* Generator wave Phase */
int       gen_freq[MAXGENERATOR];      /* Generator Frequency  */
long long gen_phaseincr[MAXGENERATOR]; /* Generator Phase Increment  */
short     *gen_table[MAXGENERATOR];    /* Generator Sample Table */
short     gen_vol[MAXGENERATOR];       /* Generator Volume setting */

short sintab[MAXSINTAB];         /* Sine table           */
short sst[MAXSST];             /* A short sample table */
int maxsampletable = 1;        /* Two built in sample tables */
short *sampletables[MAXSAMPLETABLE];

int       fil_freq[MAXGENERATOR];     /* Filter Frequency      */
short     fil_reso[MAXGENERATOR];     /* Filter Resonance      */

int midinote[MAXPOLY];
int vel[MAXPOLY];
double bend = 0.0;
double modulation = 1.0;
int volume = 64;
FILE *tclf;                    /* socket driven tcl file (ptr)       */


extern int do_read();
extern char inlin[];
extern int serv_main();
extern int tcpip_port;
char line[4*1024];


simulate(tn,p)
int tn;
short *p;
{
   int t;
   for (t=0; t<tn; t++) {
      computeoutputsample(p);  /* compute the next sample pair */
      p+=2;                    /* Is a Portable Audio-lib maintained var */
   }
}

computeoutputsample(p)
short *p;
{
   int i, t;
   long tl;
   int l, r;   /* Left and Right output samples */
   l = 0; r = 0;

   for (i=0; i<MAXGENERATOR; i++)
      if (gen_state[i] == GENRUN) {
	 t = (((int) gen_vol[16]) * ((gen_table[16])[(int) (gen_phase[16] / 1000)]) )/1000;
         tl = (( 32000 + (long) t) * gen_phaseincr[i]) / 32000;
         gen_phase[i] = (gen_phase[i] + tl) % ((long long) 1000 * (long long) MAXSINTAB);
	 t = (((int) gen_vol[i]) * ((gen_table[i])[(int) (gen_phase[i] / 1000)]) )/1000;
         l += t;
         r += t;
      }
   
   /* Buffer this sample pair for sound library interrupt routine */
   *p++ = (short) (l) ;
   *p   = (short) (r) ;
}

init_sounds()
{
   int i;
   
   for (i=0; i< MAXGENERATOR; i++) {
      gen_state[i] = GENINIT;    /* Means initialized tone, and quiet */
      gen_phase[i] = 0;          /* Phase in table offset units */
      gen_freq[i] = (long long) 100000L;      /* Frequency in Milli herz */
      gen_phaseincr[i] = 1;
      gen_phaseincr[i] *= MAXSINTAB;
      gen_phaseincr[i] *= gen_freq[i];
      gen_phaseincr[i] /= (long long) 44100L;
      gen_table[i] = (void *) &sst[0];
      gen_vol[i] = 1000;        /* Volume, range 0 - 1000 */
   }
   for (i=0; i< MAXSINTAB; i++)
      sintab[i] = (short) (32000.0 * sin(((double) i) *2.0*PI/MAXSINTAB) );
   for (i=0; i< MAXSST; i++)
      sst[i] = (short) ((32000.0/4) * (
         sin(((double) i) * 1.0 *2.0*PI/MAXSINTAB)/1.0 +
         sin(((double) i) * 2.0 *2.0*PI/MAXSINTAB)/2.0 +
         sin(((double) i) * 3.0 *2.0*PI/MAXSINTAB)/3.0 +
         sin(((double) i) * 4.0 *2.0*PI/MAXSINTAB)/4.0
	 ) );
   sampletables[0] = (short *) &sintab[0];
   sampletables[1] = (short *) &sst[0];
   maxsampletable = 1;
   gen_state[0] = GENRUN;
         gen_phase[0] = (gen_phase[0] + gen_phaseincr[0]) % ((long long) 1000 * (long long) MAXSINTAB);
/* printf("%ld\n",gen_phase[0]); */
}

process_args(argc,argv)
int argc;
char *argv[];
{
   int i;

   if ((argc<1) || (argc >2)) {
      printf("Usage %s [control TCP/IP port #, default = %d] \n",
         argv[0], tcpip_port);
      exit(-1);
   }
   if (argc > 2) {
      sscanf(argv[1],"%d",&tcpip_port);
   }
}

int val;
char name[1024];

sockproc(r)
int r;
{
   int rr,i,j;
   int b,c;
   static int count = 0;

   count+=r;

   while (  (rr = do_read()) > 0) {
      /* inlin[strlen(inlin)-2] = '\0'; */
      printf("set lastread {"); printf("%s",inlin); printf("}\n");
      fflush(stdout);
      if (strncmp(inlin,"val",strlen("val")) == 0) {
         sscanf(&inlin[strlen("val")],"%d",&val);
         printf("val = %d\n", val);
      }
      else
      if (strncmp(inlin,"cur",strlen("cur")) == 0) {
         char t[1024]; t[0] = 0;
         for (i=0; i < MAXPOLY; i++) {
            if (midinote[i] > 0) {
               sprintf(t,"noteon %d velocity %d index %d\n",midinote[i],vel[i], i);
               fprintf(stdout,t);
               fflush(stdout);
               do_write(t);
            }
         }
      } else
      if (strncmp(inlin,"genstate",strlen("genstate")) == 0) {
         sscanf(&inlin[strlen("genstate")],"%d %d", &b, &c);
         gen_state[b] = c;
      } else
      if (strncmp(inlin,"genfreq",strlen("genfreq")) == 0) {
         sscanf(&inlin[strlen("genfreq")],"%d %d", &b, &c);
         gen_freq[b] = c;
         gen_phaseincr[b] = (MAXSINTAB * (long long) gen_freq[b]) / (long long) 44100L;
 printf("%ld\n",gen_phaseincr[b]);
      } else
      if (strncmp(inlin,"genvol",strlen("genvol")) == 0) {
         sscanf(&inlin[strlen("genvol")],"%d %d", &b, &c);
         gen_vol[b] = c;
      } else
      if (strncmp(inlin,"filfreq",strlen("filfreq")) == 0) {
         sscanf(&inlin[strlen("filfreq")],"%d %d", &b, &c);
         fil_freq[b] = c;
      } else
      if (strncmp(inlin,"filreso",strlen("filreso")) == 0) {
         sscanf(&inlin[strlen("filreso")],"%d %d", &b, &c);
         fil_reso[b] = c;
      } else
      if (strncmp(inlin,"gentable",strlen("gentable")) == 0) {
         sscanf(&inlin[strlen("gentable")],"%d %d", &b, &c);
         gen_table[b] = sampletables[c];
      } else
      if (strncmp(inlin,"getsst",strlen("getsst")) == 0) {
         char t[1024];
         if (sscanf(&inlin[strlen("getsst")],"%d %d", &b, &c) > 1)
	    i = c; else i=0;
	 j = 0;
	 do_write("sst: \n");
	 while (j<1000 && i<MAXSST) {
	    sprintf(t,"%d %d ", i, (int) sst[i]);
            do_write(t);
	    j++;
	    i+=b;
            if ((j%4) == 3) do_write("\n");
	 }
	 do_write("\n<endsst>\n");
      } else
      if (strncmp(inlin,"makeadditivetable",strlen("makeadditivetable")) == 0) {
         double f[8], a[8], d;
         sscanf(&inlin[strlen("makeadditivetable")],
            "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf ", 
            &b, &f[0], &a[0], &f[1], &a[1], &f[2], &a[2], &f[3], &a[3], 
                &f[4], &a[4], &f[5], &a[5], &f[6], &a[6], &f[7], &a[7] );
         for (i=0; i< MAXSST; i++) {
            d = 0.0;
            for (j=0; j<8; j++) {
               d +=   a[j] * sintab[(int) ((i * f[j])) % MAXSINTAB ];
               /* d +=   a[j] * sin(((double) i) * f[j] *2.0*PI/MAXSINTAB); */
            }
            /* sst[i] = (short) ((32000.0) * ( d ) ) ; */
            sst[i] = (short) ( d ) ;
         }
      } else
      if (strncmp(inlin,"pitchbend",strlen("pitchbend")) == 0) {
         sscanf(&inlin[strlen("pitchbend")],"%d",&c);
         bend = ((double) c ) ;
         for (i=0; i< MAXPOLY; i++)
            if (midinote[i] > 0) {
               /* n[i] = 3+101* pow(2,-(midinote[i] -57.0-12+2*bend)/12.0); */
            }
      } else
      if (strncmp(inlin,"noteon",strlen("noteon")) == 0) {
         sscanf(&inlin[strlen("noteon")],"%d %d",&b,&c);
         i=0; while (i<MAXPOLY && midinote[i] > 0) i++;
         if (i >= MAXPOLY) i=0;
         /* n[i] = 3+101* pow(2,-(b-57.0)/12.0); */
         /*n[i] = 3+101* pow(2,-(b-57.0-12+2*bend)/12.0);*/
         vel[i] = c;
         midinote[i] = b;
      } else
      if (strncmp(inlin,"noteoff",strlen("noteoff")) == 0) {
         sscanf(&inlin[strlen("noteoff")],"%d",&b);
         /* pms_do(0);   * silence all strings */
         for (i=0; i< MAXPOLY; i++)
            if (midinote[i] == b) { 
	       midinote[i] = -1;
            }
      } else
      if (strncmp(inlin,"modulation",strlen("modulation")) == 0) {
         sscanf(&inlin[strlen("modulation")],"%d",&c);
         modulation = ((double) c);
      } else
      if (strncmp(inlin,"volume",strlen("volume")) == 0) {
         sscanf(&inlin[strlen("volume")],"%d",&c);
         volume = c;
   /*      volumechanged(); */
      }

      if (strncmp(inlin,"exit",strlen("exit")) == 0) {
         do_write("Pmsb exiting\n");
         exit(0);
      }
   }
}

init_midi()
{
   int i;

   for(i=0; i< MAXPOLY; i++) {
     midinote[i] = -1;
     vel[i] = 0;
   }

   modulation = 0.0;
   bend = 1.0;
   volume = 64;

}

main(argc, argv)
int argc;
char *argv[];
{
   process_args(argc,argv);

   init_sounds();     /* Initialize sound generators */
   serv_main();       /* call server connection startup routine */
   pa_maini();        /* start portable sound library */
   init_midi();       /* initialize midi or comparable socket interface */

   sockproc(10);      /* loop here waiting for messages, while playing sounds */

   pa_maine();        /* close sound channel */
   return 0;          /* assume server link goes away with process end .. */
}

